// BGR Implementation accompanying Boston University Tech Report
// Sequential Aggregate Signatures with Lazy Verification for S-BGP
// http://www.cs.bu.edu/~goldbe/papers/bgpsec-sigs.html
// 
// April 2, 2011
//
// Contained in this library are functions for generating and verifying BGR signatures, as well as some basic I/O functions used solely for testing and demo purposes.
// This library and demo requires a current version of the libssl and libssl-dev packages under Debian/Ubuntu, or the equivilant for other flavors of POSIX compliant operating systems.
// Example uses of the functions are contained in addsig.c and versig.c, which are simple programs to write a new signature into sig.txt in the current directory, and verify said signature.  These both rely on the public and private keys, as well as the message files, in this directory.  Usage is detailed in the README.txt file in this directory.
// To compile against the library, use gcc flags -L. and -lbgr
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/hmac.h>
#include <openssl/engine.h>
#include <openssl/rand.h>
#include <openssl/pem.h>
#include <openssl/err.h>
#include <string.h>


// These are the corresponding lengths in the paper, except divided by 8, because they are in bytes
// LX in the code is L_\pi in the paper
#define LH 32
#define LX 256
#define LR 16
#define LK 32
#define HASH_ALGORITHM EVP_sha256
#define DEBUG 0 // for printing various intermediate resutls to stdout

// Stores signatures so far and the final signature
struct BGR_sig_st{
  unsigned char **r;  // r[1] ... r[i] contain strings of length rlen; r[0] is NULL
  unsigned int rlen;
  unsigned char *b;   // b[1] ... b[i] contain a single bit each; b[0] is meaningless
  unsigned char *x;   // x is the current value x[i] of length xlen
  unsigned int xlen;
  unsigned char *h;   // h is the current value h[i] of length hlen
  unsigned int hlen;
  unsigned int i;     // number of signers who contributed to this signature so far
};

// Stores BGR private keys
// (BGR public keys are just RSA public keys, so no struct is provided for that)
struct BGR_key_st{ 
  RSA *rsa;
  unsigned char *k;
  int klen;
};

typedef struct BGR_sig_st BGR_sig;
typedef struct BGR_key_st BGR_key;

// Allocates a BGR signature big enough to hold the signature form the i-th signer
BGR_sig *BGR_sig_new(int i)
{
  int j;
  BGR_sig *bgr = (BGR_sig *)OPENSSL_malloc(sizeof(BGR_sig));
  if(bgr == NULL)
    {
      printf("Not enough memory to allocate new BGR_sig...\n");
      exit(1);
    }
  bgr->i = i;
  bgr->r = (unsigned char **)OPENSSL_malloc((i+1) * sizeof(unsigned char *));

  if(bgr->r == NULL)
    {
      printf("Not enough memory to allocate new BGR_sig...\n");
      exit(1);
    }
  bgr->r[0] = NULL; 
  for(j=1;j<=i;j++) 
  {
    bgr->r[j] = (unsigned char *)OPENSSL_malloc(LR);
    if(bgr->r[j] == NULL)
      {
	printf("Not enough memory to allocate new BGR_sig...\n");
	exit(1);
      }
  }
  
  bgr->rlen = LR;
  bgr->x = (unsigned char *)OPENSSL_malloc(LX); 
  bgr->xlen = LX;
  bgr->h = (unsigned char *)OPENSSL_malloc(LH);
  bgr->hlen = LH;
  bgr->b = (unsigned char *)OPENSSL_malloc(i+1); 
  if(bgr->x == NULL || bgr->h == NULL || bgr->b == NULL)
    {
      printf("Not enough memory to allocate new BGR_sig...\n");
      exit(1);
    }
  return bgr;
}

// Frees the elements of the BGR_sig structure and the structure itself
void BGR_sig_free(BGR_sig *sig)
{
  int j;
  for(j=1;j<=sig->i;j++)
    {
      OPENSSL_free(sig->r[j]);
    }
  OPENSSL_free(sig->r);
  OPENSSL_free(sig->x);
  OPENSSL_free(sig->h);
  OPENSSL_free(sig->b);
  OPENSSL_free(sig);
  return;
}

// Generates a BGR public/private key pair
// Assumes RAND has been initialized using RAND_seed
BGR_key *BGR_gen(int num, unsigned long e)
{
  BGR_key *bgrkey = OPENSSL_malloc(sizeof(BGR_key));
  if(bgrkey == NULL)
    {
      printf("Not enough memory to allocate new BGR_key...\n");
      exit(1);
    }
  bgrkey->rsa = RSA_generate_key(num,e,NULL,NULL); 
  bgrkey->klen = LK;
  bgrkey->k = OPENSSL_malloc(LK);
  if(bgrkey->k == NULL || bgrkey->rsa == NULL)
    {
      printf("Not enough memory to allocate new BGR_key...\n");
      OPENSSL_free(bgrkey);
      exit(1);
    }
  RAND_bytes(bgrkey->k,LK);
  return bgrkey;
}

// Creates a BGR key out of an RSA key and a PRF key that are passed in
static BGR_key *BGR_populate_key(RSA *rsa, char *k, int len)
{
  BGR_key *bgrkey = OPENSSL_malloc(sizeof(BGR_key)); 
  if(bgrkey == NULL)
    {
      printf("Not enough memory to allocate new BGR_key...\n");
      exit(1);
    }
  bgrkey->rsa = rsa;
  bgrkey->klen = len;
  bgrkey->k = k;
  return bgrkey;
}


// Creates a new BGR signature object for the next signer,
// i.e., (prev->i)+1. Populates it with
// r and b values from prev, because these values will not change
static BGR_sig *copy_grow(BGR_sig *prev)
{
  int j,k;
  BGR_sig *ret = BGR_sig_new((prev->i) + 1);
  for(j=1;j<=prev->i;j++)
    memcpy(ret->r[j],prev->r[j],LR);
  memcpy((ret->b) + 1, (prev->b) + 1, prev->i);
  memcpy(ret->x,prev->x,prev->xlen);
  memcpy(ret->h,prev->h,prev->hlen);
  return ret;
}

// Frees the elements of the BGR_key structure and the structure itself  
void BGR_key_free(BGR_key *key)
{
  RSA_free(key->rsa);
  OPENSSL_free(key->k);
  OPENSSL_free(key);
}

// places the fingerprint of the RSA public key into out;
// out needs to point to a buffer of at least outlen bytes
// if the fingerprint cannot fit into outlen bytes,
// prints an error and dies.
static void key_fingerprint(RSA *rsa, unsigned char *out, int outlen)
{
  if(outlen <  2*LX)
    {
      printf("Error: fingerprint too large for buffer\n");
      exit(1);
    }
 
#if DEBUG
  printf("2LX %d\n",2*LX);
  printf("n %d\n",BN_num_bytes(rsa->n));
  printf("e %d\n",BN_num_bytes(rsa->e));
  printf("sum %d\n",BN_num_bytes(rsa->n) + BN_num_bytes(rsa->e));

#endif
  memset(&(out[0]),'\0',LX - BN_num_bytes(rsa->n));
  BN_bn2bin(rsa->n,&(out[LX - BN_num_bytes(rsa->n)]));
  memset(&(out[LX]),'\0',LX - BN_num_bytes(rsa->e));
  BN_bn2bin(rsa->e,&(out[LX + LX - BN_num_bytes(rsa->e)]));
#if DEBUG
  for(i=0;i<2*LX;i++)
    printf("%.2x ",out[i]);
#endif
}



// Internal signature function
// Comments referencing line numbers correspond to Algorithm 9 in the tech report.
static void BGR_sign_internal(unsigned char *mi, int mlen, BGR_key *bgr_key,char *fprint, int fprintlen, BGR_sig *sigret)
{
  //Line 1
  if((sigret->xlen != LX) || (sigret->hlen != LH)) // TBD: once we get away from #defines, this will need to check that it matches the signer's idea of the what the lengths should be (note that the RSA modulus should also reflect that idea); perhaps have a BGR_CTX that has to be intialized by the user with the correct lengths.  
    {
      printf("Error: sign fail; lengths don't match.\n");
    }

  HMAC_CTX ctx;
  EVP_MD_CTX mdctx;
  unsigned int eta_len, hmaclen;
  int i,enc;
  int EVP_status = 1;
  unsigned char *eta = (unsigned char *)OPENSSL_malloc(LK);
  unsigned char *g = (unsigned char *)OPENSSL_malloc(LX);
  unsigned char *r = (unsigned char *)OPENSSL_malloc(EVP_MD_size(HASH_ALGORITHM()));
  if(eta == NULL || g == NULL || r == NULL)
    {
      printf("Not enough memory to generate signature.\n");
      exit(1);
    }
 
  //Line 3
  HMAC_CTX_init(&ctx);
  HMAC_Init(&ctx, bgr_key->k, bgr_key->klen, HASH_ALGORITHM());
  HMAC_Update(&ctx,mi,mlen);
  HMAC_Update(&ctx,sigret->x, sigret->xlen ); 
  HMAC_Update(&ctx,sigret->h, sigret->hlen );
  HMAC_Final(&ctx,r,&hmaclen); //ith r for signing, but -1 for zero index
  HMAC_CTX_cleanup(&ctx);
  
  if(sigret->rlen > EVP_MD_size(HASH_ALGORITHM()))
    {
      printf("Not enough HMAC output to cover all of r.\n");
      exit(1);
    }
  memcpy(sigret->r[sigret->i],r,sigret->rlen);  //The HMAC hash r is being truncated to rlen bits and stored in the signature object 
#if DEBUG
  printf("rlen:\t%u\n",sigret->rlen);
  printf("R OUTPUT:\n");
  for(i=0;i<sigret->rlen;i++)
    printf("%.2x ",sigret->r[sigret->i][i]);
  printf("END R\n\n");
#endif
#if DEBUG
  printf("fprint OUTPUT:\n");
  for(i=0;i<fprintlen;i++)
    printf("%.2x ",(unsigned char)fprint[i]);
  printf("END R\n\n");
#endif

  //Line 6
  EVP_MD_CTX_init(&mdctx);
  if(sigret->hlen != EVP_MD_size(HASH_ALGORITHM()))
    {
      printf("Hash length of signature object does not match hash length of SHA256\n");
      exit(1);
    }
  EVP_status &= EVP_DigestInit_ex(&mdctx, HASH_ALGORITHM(), NULL); 
  EVP_status &= EVP_DigestUpdate(&mdctx, fprint, fprintlen);
  EVP_status &= EVP_DigestUpdate(&mdctx, mi, mlen);
  EVP_status &= EVP_DigestUpdate(&mdctx, sigret->x, sigret->xlen);
  EVP_status &= EVP_DigestUpdate(&mdctx, sigret->r[(sigret->i)], sigret->rlen);
  EVP_status &= EVP_DigestFinal_ex(&mdctx, eta, &eta_len);
  EVP_MD_CTX_cleanup(&mdctx);
  if(EVP_status == 0) //If this occurs, one of the EVP_Digest functions failed
      {
	printf("Signature generation failed due to failure of EVP operation.\n");
	exit(1);
      }
   if(eta_len != sigret->hlen)
    {
      printf("Hash function produced improperly sized output!\n");
      exit(1);
    }
#if DEBUG
  printf("ETA OUTPUT:\n");
  for(i=0;i<eta_len;i++)
    printf("%.2x ",eta[i]);
  printf("END ETA\n\n");
#endif
  //Line 7
  for (i = 0; i < sigret->hlen; i++) 
    sigret->h[i] = eta[i] ^ sigret->h[i];

#if DEBUG
  printf("hi OUTPUT:\n");
  for(i=0;i<LH;i++)
    printf("%.2x ",sigret->h[i]);
  printf("END hi\n\n");
#endif
  //Line 8
  if (PKCS1_MGF1(g,sigret->xlen,sigret->h,sigret->hlen,HASH_ALGORITHM()) < 0) 
    {
      printf("Error in MGF1 function\n");
      exit(1);
    }
  //Line 9
  for (i = 0; i < sigret->xlen; i++) 
    g[i] ^= (sigret->x)[i]; 
  //Line 10
  g[0] &= 0x7F; //zero first bit
  //Line 11 is not needed, since BS2IP is handled internally inside RSA_private_encrypt
  //Line 12 
  enc = RSA_private_encrypt(sigret->xlen,g,sigret->x,bgr_key->rsa,RSA_NO_PADDING); 
  if(enc != sigret->xlen)  
    {
      printf("Problem with RSA sign for xi\n");
      exit(1);
    }
  
  //Line 13 is not needed since I2BSP is handled internally inside RSA_private_encrypt
  //Line 14 
  sigret->b[sigret->i] = ((sigret->x[0]) & 0x80) >> 7 ; // Store the most significant bit of x in b.

  //Line 15
  sigret->x[0] &= 0x7F; // Zero out the first bit of x

  OPENSSL_free(eta);
  OPENSSL_free(g);
  OPENSSL_free(r);
}

//Internal signature function
//Comments referencing line numbers correspond to Algorithm 10 in the tech report.
//Returns 0 if verification fails, and 1 if verification succeeds.
static int BGR_verify_internal(RSA **public, char **messages, int *mlen, char **fprint, int fprintlen, BGR_sig *sig)
{
  unsigned int i, eta_len;
  int result,j;
  int EVP_status = 1;
  EVP_MD_CTX mdctx;
  unsigned char *xcur = (unsigned char *) malloc(sig->xlen); // These aren't OPENSSL_mallocs because the material is all "public" data.
  unsigned char *hcur = (unsigned char *) malloc(sig->hlen);
  unsigned char *yi = (unsigned char *) malloc(sig->xlen);
  unsigned char *etai = (unsigned char *) malloc(sig->hlen);
  if(xcur == NULL || hcur == NULL || yi == NULL || etai == NULL)
    {
      printf("Not enough memory to verify signature...\n");
      exit(1);
    }
  memcpy(xcur,sig->x,sig->xlen);
  memcpy(hcur,sig->h,sig->hlen);
  for(i=sig->i;i>0;i--)
    {

#if DEBUG
      printf("%d\n",i);
      printf("XCUR OUTPUT:\n");
      for(j=0;j<256;j++)
	printf("%.2x ",xcur[j]);
      printf("END XCUR\n\n");
#endif
#if DEBUG
  printf("fprint OUTPUT:\n");
  for(j=0;j<fprintlen;j++)
    printf("%.2x ",(unsigned char)fprint[1][j]);
  printf("END R\n\n");
#endif
      //Step 2: set most significant bit of xcur to b[i]
      xcur[0] |= (sig->b[i] << 7);
      //Steps 3 through 5 in the implementation are contained within the RSA_public_decrypt function.
      //Step 6: RSA Verification
      result = RSA_public_decrypt(sig->xlen,xcur,yi,public[i],RSA_NO_PADDING);

#if DEBUG
  printf("rlen:\t%u\n",sig->rlen);
  printf("R OUTPUT:\n");
  for(j=0;j<sig->rlen;j++)
    printf("%.2x ",sig->r[i][j]);
  printf("R HMAC\n\n");
#endif
      if(result == -1)
	{
	  free(xcur);
	  free(hcur);
	  free(yi);
	  free(etai);
	  return 0;
	}
      //Step 7 in the implementation is contained within the RSA_public_decrypt function.
      //Step 8
      //Reuse xcur for gi, since xcur no longer needed
      if(PKCS1_MGF1(xcur,sig->xlen,hcur,sig->hlen,HASH_ALGORITHM()) < 0)  //If MGR1 function fails
	{
	  free(xcur);
	  free(hcur);
	  free(yi);
	  free(etai);
	  return 0;
	}
      //Step 9
      for(j=0;j<sig->xlen;j++)
	xcur[j] ^= yi[j];
      xcur[0] &= 0x7F;
#if DEBUG
      printf("XCUR-1 OUTPUT:\n");
      for(j=0;j<sig->xlen;j++)
	printf("%.2x ",xcur[j]);
      printf("END XCUR-1\n\n");
#endif

      //Step 10
      EVP_MD_CTX_init(&mdctx);
      EVP_status &= EVP_DigestInit_ex(&mdctx, HASH_ALGORITHM(), NULL);
      EVP_status &= EVP_DigestUpdate(&mdctx, fprint[i], fprintlen);
      EVP_status &= EVP_DigestUpdate(&mdctx, messages[i], mlen[i]);
      EVP_status &= EVP_DigestUpdate(&mdctx, xcur, sig->xlen);
      EVP_status &= EVP_DigestUpdate(&mdctx, sig->r[i], sig->rlen);
      EVP_status &= EVP_DigestFinal_ex(&mdctx, etai, &eta_len); 
      EVP_MD_CTX_cleanup(&mdctx);
      if(EVP_status == 0)//If this occurs, one of the EVP_Digest functions failed
	{
	  printf("Verification failed due to EVP Digest failure.\n");
	  exit(1);
	}
      if(eta_len != sig->hlen)
	{
	  printf("Hash function produced improperly sized output!\n");
	  exit(1);
	}
#if DEBUG
      printf("ETA OUTPUT:\n");
      for(j=0;j<sig->hlen;j++)
	printf("%.2x ",etai[j]);
      printf("END ETA\n\n");
#endif
      //Step 11
      for(j=0;j<sig->hlen;j++)
	hcur[j] ^= etai[j];
#if DEBUG
  printf("hi OUTPUT:\n");
  for(j=0;j<sig->hlen;j++)
    printf("%.2x ",hcur[j]);
  printf("END hi\n\n");
#endif

    }  //End of the main loop
 
  free(etai);
  free(yi);
  

  int ret = 1;
  for (i=0; i<sig->xlen; i++)
    if (xcur[i]!='\0') {
	  ret = 0;
	  break;
	}
  if (ret==1) {
    for (i=0; i<sig->hlen; i++)
      if (hcur[i]!='\0') {
	ret = 0;
	break;
      }
  }
  free(xcur);
  free(hcur);
  return ret;

  
}


BGR_sig *BGR_sign(unsigned char *mi, int mlen, BGR_key *bgr_key, BGR_sig *prevsign)
{
  int nullprev = 0;
  int fprintlen = 2*LX;
  BGR_sig *newsig;
  unsigned char *fprintbuf = OPENSSL_malloc(fprintlen);
  if(fprintbuf == NULL)
    {
      printf("Not enough memory to generate signature...\n");
      exit(1);
    }
  key_fingerprint(bgr_key->rsa,fprintbuf,fprintlen);
  if(prevsign == NULL)
    {
      nullprev = 1;
      newsig = BGR_sig_new(1);
      memset(newsig->h,'\0',newsig->hlen);
      memset(newsig->x,'\0',newsig->xlen);
      
    }
  else
    {
      newsig = copy_grow(prevsign);
    }
  BGR_sign_internal(mi, mlen, bgr_key, fprintbuf, fprintlen, newsig); 
  OPENSSL_free(fprintbuf);
  return newsig;
}

//public is an array of the public keys of signers who have added to the signature
//messages is an array of the messages signed at each step
//mlen is an array of message lengths (corresponding to array messages)
//sig is the signature object itself
int BGR_verify(RSA **public, char **messages, int *mlen, BGR_sig *sig)
{
  int fprintlen = 2*LX;
  char **fprintarray = malloc((sig->i + 1) * (sizeof (char *))); // No secret data here, hence malloc and not OPENSSL_malloc
  int i;
  if(fprintarray == NULL)
    {
      printf("Not enough memory for verification.\n");
      exit(1);
    }
  for(i=1;i<=sig->i;i++)
    {
      fprintarray[i] = malloc(fprintlen);  // No secret data here, hence malloc and not OPENSSL_malloc
      if(fprintarray[i] == NULL)
	{
	  printf("Not enough memory for verification.\n");
	  exit(1);
	}
      key_fingerprint(public[i],(unsigned char *)fprintarray[i],fprintlen);
    }
  int result = BGR_verify_internal(public, messages, mlen, fprintarray, fprintlen, sig);
  for(i=1;i<=sig->i;i++)
    free(fprintarray[i]);
  free(fprintarray);
  return result;
}

//******************************************************************
// ONLY I/O FUNCTIONALITY BELOW THIS LINE
//******************************************************************/

int hexDigitToInt(char c)
{
  if (c>='0' && c<='9')
    return c-'0';
  if (c>='a' && c<='f')
    return c-'a'+10;
  if (c>='A' && c<='F')
    return c-'A'+10;
  return 0;
}

void unsignedCharStarToHex(unsigned char * s, int len, FILE * fp)
{
  int i;
  
  for (i=0; i<len; i++)
    fprintf(fp, "%.2x", s[i]);
}

void hexToUnsignedCharStar(unsigned char * s, int len, FILE * fp) 
{
  int i;
  
  for (i=0; i<len; i++) {
    s[i]=hexDigitToInt(fgetc(fp))*16;
    s[i]+=hexDigitToInt(fgetc(fp));
  }
}
    

void sigToFile(BGR_sig * sig)
// always writes to sig.txt
{
  int i;
  FILE * fp=fopen("sig.txt", "wt");
  
  if (fp == NULL) {
    printf("Error: Cannot open signature file for writing!\n");
    exit(1);
  }
#if DEBUG
  printf("%d", sig->i);
#endif
  for(i=1; i<=sig->i; i++) {
    fprintf(fp, "%d",sig->b[i]);
  }
  fprintf(fp, "\n\n");
  for(i=1; i<=sig->i; i++)
    unsignedCharStarToHex(sig->r[i], sig->rlen, fp);
  fprintf(fp, "\n\n");
  unsignedCharStarToHex(sig->x, sig->xlen, fp);
  fprintf(fp, "\n\n");
  unsignedCharStarToHex(sig->h, sig->hlen, fp);
  fprintf(fp, "\n");
    
  fclose (fp);
}

BGR_sig * sigFromFile(int signerNumber)
// always reads from sig.txt, but checks that signerNumber matches the length of the signature sig.txt contains
// skips the check if signerNumber is -1
// CANNOT read 0th signature -- such a thing does not exist
{
  int i;
  BGR_sig* sig;
  char temp;
  FILE * fp = fopen("sig.txt", "rt");
  
  if (fp == NULL) {
    printf("Error: Cannot open signature file for reading!\n");
    exit(1);
  }
    
  // count the length of the b string -- that will tell you the correct value for signerNumber
  for (i=0; ;i++) {
    temp = fgetc(fp);
    if (temp == '\n' || temp == '\r')
      break;
  }

  if (signerNumber!=-1 && i!=signerNumber) {
    printf("sig.txt contains a signature by %d signers, not %d signers.\n", i, signerNumber);
    fclose (fp);
    exit(1);
  }
  fseek(fp, 0, SEEK_SET); // note: relative seek back -i-1 characters would accomlish the same, but screws up on Windows in text mode because of \n\r issues

  sig = BGR_sig_new (i);

  for(i=1; i<=sig->i; i++) {
    if(fgetc(fp)=='1')
      sig->b[i]=1;
    else
      sig->b[i]=0;
  }
  
  // skip the new line
  do {
    temp = fgetc (fp);
  } while (temp == '\n' || temp == '\r'); 
  ungetc(temp, fp);
  
  for(i=1; i<=sig->i; i++)
    hexToUnsignedCharStar(sig->r[i], sig->rlen, fp);

  // skip the new line
  do {
    temp = fgetc (fp);
  } while (temp == '\n' || temp == '\r'); 
  ungetc(temp, fp);

  hexToUnsignedCharStar(sig->x, sig->xlen, fp);

  // skip the new line
  do {
    temp = fgetc (fp);
  } while (temp == '\n' || temp == '\r'); 
  ungetc(temp, fp);

  hexToUnsignedCharStar(sig->h, sig->hlen, fp);
  fclose (fp);
  return sig;
}

BGR_key * privateKeyFromFile (int signerNumber) {
// reads key from file private_key_signerNumber.txt 
// The key should start on line one with LK-length hex string of k and then contain the private key below
  char filename[50];
  FILE * fp;
  BGR_key *bgrkey;
  
  sprintf(filename, "private_key_%d.txt", signerNumber);
  fp=fopen(filename, "rt");
  if (fp == NULL) {
    printf("Error: cannot open private key file %s for reading\n", filename);
    exit(1);
  }
  
  bgrkey = OPENSSL_malloc(sizeof(BGR_key));
  bgrkey->klen = LK;
  bgrkey->k = (unsigned char*) OPENSSL_malloc(bgrkey->klen);
  hexToUnsignedCharStar (bgrkey->k, bgrkey->klen, fp);
  fseek(fp,0,SEEK_SET);
  if (!PEM_read_RSAPrivateKey(fp,&(bgrkey->rsa),NULL,NULL))  {
    printf("Error loading RSA Private Key File %s.\n", filename);
    exit (1);
  }
  fclose(fp);
  return bgrkey;
}

RSA * singlePublicKeyFromFile(int signerNumber) {
  char filename[50];
  FILE * fp;
  RSA * ret = RSA_new();
  
  sprintf(filename, "public_key_%d.txt", signerNumber);
  fp=fopen(filename, "rt");
  if (fp == NULL) {
    printf("Error: cannot open public key file %s for reading\n", filename);
    exit(1);
  }
  if (!PEM_read_RSA_PUBKEY(fp,&ret,NULL,NULL))  {
    printf("Error loading RSA Public Key File %s.\n", filename);
    exit (1);
  }
  fclose(fp);
  return ret;
}

RSA ** publicKeysFromFile (int signerNumber)
// return a vector of signerNumber+1 keys, 0th one being NULL
// reads keys from files public_key_1.txt, public_key_2.txt, ..., public_key_signerNumber.txt
{
  int i;
  RSA ** ret = OPENSSL_malloc((signerNumber+1)*sizeof(RSA*));
  ret[0]=NULL;
  for (i = 1; i<=signerNumber; i++)
    ret[i]=singlePublicKeyFromFile(i);
  return ret;
}

// return a the message from msg_signerNumber.txt
// it is placed into *message; its length is placed into *length
// If something was in *message before it WILL NOT be freed
void singleMessageFromFile (int signerNumber, unsigned char **message, int *length)
{
  char filename[50];
  FILE * fp;
  int i;
     
  sprintf(filename, "msg_%d.txt", signerNumber);
  fp=fopen(filename, "rb");
  if (fp == NULL) {
    printf("Error: cannot open message file %s for reading\n", filename);
    exit(1);
  }
  
  fseek(fp, 0, SEEK_END); // The easiest way to get the length is to seek to the end
  *length = ftell(fp);
  *message = (unsigned char *) OPENSSL_malloc (*length);
  fseek(fp, 0, SEEK_SET);
  for (i = 0; i<*length; i++)
    (*message)[i]=fgetc(fp);
  fclose(fp);
 }
  
  
// return a vector of signerNumber+1 messages, 0th one being NULL
// reads keys from files msg_1.txt, msg_2.txt, ..., msg_signerNumber.txt
// The vector of messages is placed into *messages; the vector of lengths into *lengths
// If something was in *messages and *lengths before, it WILL NOT be freed
void messagesFromFile (int signerNumber, unsigned char *** messages, int ** lengths)
{
  int i;
  *messages = (unsigned char **)OPENSSL_malloc((signerNumber+1)*sizeof(char*));
  *lengths = (int *) OPENSSL_malloc((signerNumber+1)*sizeof(int));
  (*messages)[0] = NULL;
  (*lengths)[0] = 0;
  for (i = 1; i<=signerNumber; i++)
    singleMessageFromFile(i, (*messages)+i, (*lengths)+i);
}
